import java.io.*;
import java.util.*;

import org.jdom2.*;
import org.jdom2.filter.Filters;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.XMLOutputter;
import org.jdom2.xpath.*;



/**
 * Takes a featuremodel.xml file generated using FeatureModelConfigurator and a given 
 * ROS/ Gazebo .gazebo.xacro file.
 * The .gazebo.xacro file is modified to support the features specified in the
 * featuremodel.xml file.
 */
public class Adapter {
	
	/**
	 * Reads an .xml file from file path and returns a JDOM document.
	 * 
	 * @param src			The .xml file path
	 * @return JDOM document
	 * @throws JDOMException
	 * @throws IOException
	 */
	public static Document readXml(String src) throws JDOMException, IOException {
		File file = new File(src);
		SAXBuilder saxBuilder = new SAXBuilder();
		Document doc = saxBuilder.build(file);

		return doc;
	}
	
	/**
	 * Executes given XPath expression for elements.
	 * 
	 * @param doc			The JDOM Document
	 * @param xPath			XPath expression
	 * @return List with all elements found or empty list if no match found
	 */
	public static List<Element> xQueryElements(Document doc, String xPath) {
		XPathFactory xFactory = XPathFactory.instance();
		XPathExpression<Element> expr = xFactory.compile(xPath, Filters.element());
		List<Element> result = expr.evaluate(doc);
		
		return result;
	}
	
	/**
	 * Executes given XPath expression for attributes.
	 * 
	 * @param doc			The JDOM Document
	 * @param xPath			XPath expression
	 * @return List with all attributes found or empty list if no match found
	 */
	public static List<Attribute> xQueryAttributes(Document doc, String xPath) {
		XPathFactory xFactory = XPathFactory.instance();
		XPathExpression<Attribute> expr = xFactory.compile(xPath, Filters.attribute());
		List<Attribute> result = expr.evaluate(doc);
		
		return result;
	}
	
	/**
	 * Searches a document for the first occurrence of an xml tag.
	 * 
	 * @param doc			The JDOM Document
	 * @param elementName	Name of the xml tag
	 * @return First element found or NULL if no match found
	 */
	public static Element searchFirstElement(Document doc, String elementName) {
		//XPath index starts with 1; [] has a higher precedence than //
		List<Element> query= xQueryElements(doc, "(//"+elementName+")[1]");
		
		if(query.isEmpty()) return null;
		return query.get(0);
	}
	
	/**
	 * Searches a document for all occurrences of an xml tag.
	 * 
	 * @param doc			The JDOM Document
	 * @param elementName	Name of the xml tag
	 * @return List with all elements found or empty list if no match found
	 */
	public static List<Element> searchAllElements(Document doc, String elementName) {
		return xQueryElements(doc, "//"+elementName);
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			Document FeatureModel = readXml("resources/input/featuremodel.xml");
			Document GazeboModel = readXml("resources/input/turtlebot3-master/turtlebot3_description/urdf/turtlebot3_burger.gazebo.xacro");
			
			//TODO
			//######### Operate on featuremodel.xml #########
			Element fmRoot = FeatureModel.getRootElement();
			List<Element> modules = fmRoot.getChildren("module");
			
			//###############################################
			
			
			//TODO
			//######### Operate on .gazebo.xacro ############
			Element gmRoot = GazeboModel.getRootElement();

			//###############################################
			
			
			FileOutputStream dest = new FileOutputStream("resources/output/turtlebot3_burger.gazebo.xacro");
			XMLOutputter xmlOutput = new XMLOutputter();
			
			//TODO
			//######### Output result #######################
			xmlOutput.output(GazeboModel, dest);

			//###############################################
			
			
			//XPathExamples
			XPathFactory xFactory = XPathFactory.instance();
			XPathExpression<Element> xExpr = xFactory.compile("robot/gazebo/sensor/ray/range/min", Filters.element());
			Element test = xExpr.evaluateFirst(GazeboModel);
			List<Element> test2 = searchAllElements(GazeboModel, "min");
			List<Element> test3 = searchAllElements(GazeboModel, "max");
			Element test4 = searchFirstElement(GazeboModel, "plugin");
			List<Attribute> test5 = xQueryAttributes(GazeboModel, "(//plugin)[1]/@*");
			
			if(test != null) xmlOutput.output(test, System.out);
			System.out.println("\n");
			xmlOutput.output(test2, System.out);
			System.out.println("\n");
			xmlOutput.output(test3, System.out);
			System.out.println("\n");
			if(test4 != null) xmlOutput.output(test4, System.out);
			System.out.println("\n");
			for(Attribute a : test5) System.out.println(a.getName()+"="+a.getValue());
			
		} catch (JDOMException | IOException e) {
			e.printStackTrace();
		}
	}
}
